import warnings
import os
import importlib.util
from typing import List, Union, Iterable, Tuple
import numpy as np
from copy import deepcopy

from QuICT.core import Circuit
from QuICT.core.gate import BasicGate, Measure, GateType
from QuICT.core.operator import Trigger


cur_path = os.path.dirname(os.path.abspath(__file__))

mod_name = "sim_back_bind"
mod_path = "sim_back_bind"

for file in os.listdir(cur_path):
    if file.startswith(mod_path):
        mod_path = os.path.join(cur_path, file)

spec = importlib.util.spec_from_file_location(mod_name, mod_path)
sim_back_bind = importlib.util.module_from_spec(spec)
spec.loader.exec_module(sim_back_bind)


def get_gate_qubit_pos(gate: BasicGate) -> List[int]:
    args = deepcopy(gate.cargs)
    args.extend(deepcopy(gate.targs))
    return args


class GateDescription:
    """An interface used for type hints. This class is actually implemented in C++ side.
    """

    def __init__(
        self, gate_name: str, affect_args: Iterable[int], data_ptr: Iterable[complex]
    ):
        pass


# sim_back_bind.GateDescription: GateDescription.__class__

special_x = (GateType.x,)
special_h = (GateType.h,)
diag_1 = (
    GateType.s,
    GateType.sdg,
    GateType.z,
    GateType.id,
    GateType.u1,
    GateType.rz,
    GateType.t,
    GateType.tdg,
    GateType.phase,  # TODO: optimize this with special_phase
    GateType.gphase,
)

# The unitary_x category is some gates treated as unitary.
# The real unitary gates are not included.
unitary_1 = (
    GateType.hy,
    GateType.y,
    GateType.sx,
    GateType.sy,
    GateType.sw,
    GateType.u2,
    GateType.u3,
    GateType.rx,
    GateType.ry,
)

diag_2 = (GateType.rzz,)

unitary_2 = (
    GateType.fsim,
    GateType.rxx,
    GateType.ryy,
    GateType.rzx,
    GateType.swap,  # Maybe this could be optimized
    GateType.iswap,
    GateType.iswapdg,
    GateType.sqiswap
)

ctrl_diag = (
    GateType.cz,
    GateType.crz,
    GateType.cu1,
)

ctrl_unitary = (
    GateType.cx,  # TODO: Optimize this with special_cx
    GateType.cy,
    GateType.ch,
    GateType.cu3,
)

measure_gate = (GateType.measure,)


def gate_to_desc(gate: BasicGate) -> List[GateDescription]:
    """Helper function to create GateDescription from a quantum gate.

    Parameters
    ----------
    gate:
        A quantum gate.

    Returns
    -------
    Simple GateDescription for input gate.
    """
    gate_type = gate.type
    if gate_type in special_x:
        return [
            sim_back_bind.GateDescription(
                "special_x", list(get_gate_qubit_pos(gate)), list([])
            )
        ]
    elif gate_type in special_h:
        return [
            sim_back_bind.GateDescription(
                "special_h", list(get_gate_qubit_pos(gate)), list([])
            )
        ]
    elif gate_type in diag_1:
        return [
            sim_back_bind.GateDescription(
                "diag_1", list(get_gate_qubit_pos(gate)), list(np.diag(gate.matrix))
            )
        ]
    elif gate_type in diag_2:
        return [
            sim_back_bind.GateDescription(
                "diag_2", list(get_gate_qubit_pos(gate)), list(np.diag(gate.matrix))
            )
        ]
    elif gate_type in unitary_1:
        return [
            sim_back_bind.GateDescription(
                "unitary_1", list(get_gate_qubit_pos(gate)), list(gate.matrix.flatten())
            )
        ]
    elif gate_type in unitary_2:
        return [
            sim_back_bind.GateDescription(
                "unitary_2", list(get_gate_qubit_pos(gate)), list(gate.matrix.flatten())
            )
        ]
    elif gate_type in ctrl_diag:
        return [
            sim_back_bind.GateDescription(
                "ctrl_diag",
                list(get_gate_qubit_pos(gate)),
                list(np.diag(gate.matrix)[2:].copy()),
            )
        ]
    elif gate_type in ctrl_unitary:
        return [
            sim_back_bind.GateDescription(
                "ctrl_unitary",
                list(get_gate_qubit_pos(gate)),
                list(gate.matrix[2:, 2:].copy().flatten()),
            )
        ]
    elif gate_type in measure_gate:
        return [
            sim_back_bind.GateDescription(
                "measure", list(get_gate_qubit_pos(gate)), list([])
            )
        ]
    elif gate_type == GateType.unitary and gate.targets <= 2:
        return [
            sim_back_bind.GateDescription(
                f"unitary_{gate.targets}",
                list(get_gate_qubit_pos(gate)),
                list(gate.matrix.flatten()),
            )
        ]
    elif hasattr(gate, "build_gate"):
        # Try build gate to simpler gates
        result = []
        cgate = gate.build_gate()
        for simple_gate in cgate.gates:
            result.extend(gate_to_desc(simple_gate))
        return result
    else:
        raise NotImplementedError(f"No implementation for {gate.name}")


class CircuitSimulator:
    """ An interface used for type hints. This class is actually implemented in C++ side.

    Args:
        precision (str): The precision for the state vector, one of [single, double]. Defaults to "double".
    """

    __PRECISION = ["single", "double"]

    def __init__(self, precision: str = "double"):
        if precision not in self.__PRECISION:
            raise ValueError("Wrong precision. Please use one of [single, double].")

        self._precision = np.complex128 if precision == "double" else np.complex64
        self._circuit = None
        self._instance = sim_back_bind.CircuitSimulator()
        self._gate_desc_vec: List[GateDescription] = []
        self._pending_gates: List[BasicGate] = []

    def name(self) -> str:
        """

        Returns
        -------
        The name of circuit simulator.
        """
        return self._instance.name()

    @classmethod
    def _map_measure(cls, circuit: Circuit, measure_raw: List[int]) -> List[List[int]]:
        mid_map = []
        for gate in circuit.gates:
            if isinstance(gate, BasicGate) and gate.type == GateType.measure:
                mid_map.append(gate.targ)

        for idx, elem in enumerate(mid_map):
            circuit.qubits[elem].measured = measure_raw[idx]

    def apply_gate(self, gate: BasicGate):
        self._gate_desc_vec.append(deepcopy(gate))

    def _run(
        self, circuit: Union[Circuit, None], use_previous: bool = False
    ) -> np.ndarray:
        """Run simulation by gate description sequence and return measure gate results.

        Parameters
        ----------
        circuit:
            Quantum circuit to be simulated. If `None` is passed, then all previous gates added by
            `apply_gate` would be executed.
        use_previous:
            Start simulation on previous result
        """
        warnings.warn(
            message="Attention! You are using a working-in-process version of circuit simulator!",
            category=Warning,
            stacklevel=1,
        )
        if circuit:
            qubits = circuit.width()
            gate_set = circuit.gates

            gate_desc_vec: List[GateDescription] = []
            idx = 0
            while idx < len(gate_set):
                gate = gate_set[idx]
                idx += 1
                if isinstance(gate, Trigger):
                    for targ in gate.targs:
                        mgate = Measure & targ
                        gate_desc_vec.extend(gate_to_desc(mgate))

                    _, measure_raw = self._instance.run(
                        qubits, gate_desc_vec, use_previous
                    )
                    gate_desc_vec = []
                    measured_state, use_previous, targ_idx = 0, True, 0
                    for mstate in measure_raw:
                        measured_state << 1
                        measured_state += mstate
                        circuit[gate.targs[targ_idx]].measured = mstate
                        targ_idx += 1

                    cgate = gate.mapping(measured_state)
                    if cgate is not None:
                        cp = cgate.checkpoint
                        position = idx if cp is None else circuit.find_position(cp)
                        gate_set = (
                            gate_set[:position]
                            + deepcopy(cgate.gates)
                            + gate_set[position:]
                        )
                else:
                    gate_desc_vec.extend(gate_to_desc(gate))

        self._gate_desc_vec = gate_desc_vec

        amplitude, measure_raw = self._instance.run(
            circuit.width(), self._gate_desc_vec, use_previous
        )
        self._map_measure(circuit, measure_raw)
        self._gate_desc_vec.clear()
        if self._precision == np.complex64:
            return amplitude.astype(np.complex64)

        return amplitude

    def run(
        self,
        circuit: Union[Circuit, None],
        state_vector: np.ndarray = None,
        use_previous: bool = False
    ) -> Union[np.ndarray, Tuple[np.ndarray, List[List[int]]]]:
        """Run simulation by gate description sequence and return measure gate results.

        Parameters
        ----------
        circuit:
            Quantum circuit to be simulated. If `None` is passed, then all previous gates added by
            `apply_gate` would be executed.
        use_previous:
            Start simulation on previous result
        """
        amplitude = self._run(circuit, use_previous)
        self._circuit = circuit

        return amplitude

    def sample(self, shots: int = 1) -> List[int]:
        assert self._circuit is not None

        state_list = [0] * (1 << self._circuit.width())
        for _ in range(shots):
            # C++ simulator will automatically copy a state vector for sampling.
            # Sample operation never affects the state vector itself.
            measure_raw = self._instance.sample(self._circuit.width())
            state = 0
            for idx, measured in enumerate(measure_raw):
                measured <<= self._circuit.width() - 1 - idx
                state += measured

            state_list[state] += 1

        return state_list



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/08/13 14:30
# @Author  : Xiaoquan Xu
# @File    : polynomial_operator.py

"""
A PolynomialOperator of symbolic operators.
The superClass of FermionOperator and QubitOperator.
"""

import copy


class PolynomialOperator(object):
    """
    A polynomial of symbolic operators.
    The superClass of FermionOperator and QubitOperator.

    In this class, the operator could be represented as below.
    The first parameter in a binary tuple represents the target, while the second one represents the type.
    Two monomials can be merged if and only if their list formats only differ in the coefficient.
    The standardlization for a monomial should be accomplished in __init__ according to specific rules.

    For example, list
    [[[(i, 1), (j, 1), (k, 0), (l, 0)], 1.2], [[(k, 1), (l, 1), (i, 0), (j, 0)], -1.2], ...]
    stands for '1.2 a_i^(1) a_j^(1) a_k^(0) a_l^(0) - 1.2 a_k^(1) a_l^(1) a_i^(0) a_j^(0) + ...',

    In the following descriptions, the above list is called list format,
    while the above string is called string format.
    """
    def __init__(self, monomial=None, coefficient=1.):
        """
        Create a monomial of operators with the two given formats.
        PolynomialOperator() means zero operator, recognized by 'monomial == None'
        QubitOperator('X1') or FermionOperator('1^') or PolynomialOperator([(1,1)]) implies the coefficient is 1
        PolynomialOperator(0) or replacing 0 with some other constant is illegal

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        """
        if monomial is None or coefficient == 0:       # zero operator
            self.operators = []
            return
        if isinstance(monomial, list):
            variables = copy.deepcopy(monomial)
        elif isinstance(monomial, str):
            variables = []
            for var in monomial.split():
                variables.append(self.analyze_single(var))
        else:
            raise Exception("Illegal type of monomial.")
        self.operators = [[variables, coefficient]]

    @classmethod
    def get_polynomial(cls, monomial=None, coefficient=1.):
        '''
        Construct an instance of the same class using the arguments.
        To be overrided.

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        '''
        raise Exception("Construction of PolynomialOperator is prohibited.")

    @classmethod
    def analyze_single(cls, single_operator):
        """
        Transform a string format of a single operator to a tuple.
        To be overrided.

        Args:
            single_operator(str): string format

        Returns:
            tuple: the corresponding tuple in list format
        """
        raise Exception("The string format is not recognized.")

    @classmethod
    def parse_single(cls, single_operator):
        """
        Transform a tuple format of a single operator to a string.
        To be overrided.

        Args:
            single_operator(tuple): list format

        Returns:
            string: the corresponding string format
        """
        raise Exception("The string format is not realized.")

    def __add__(self, other):
        """
        Addition of two operators.

        Args:
            other(PolynomialOperator): Operator to be added

        Returns:
            PolynomialOperator: self + other
        """
        ans = self.get_polynomial()
        A = self.operators
        B = other.operators
        ia = ib = 0
        while ia < len(A) and ib < len(B):
            if A[ia][0] == B[ib][0]:
                if A[ia][1] + B[ib][1] != 0:
                    ans.operators += [[copy.deepcopy(A[ia][0]), A[ia][1] + B[ib][1]]]
                ia += 1
                ib += 1
            elif A[ia][0] < B[ib][0]:
                ans.operators += [copy.deepcopy(A[ia])]
                ia += 1
            else:
                ans.operators += [copy.deepcopy(B[ib])]
                ib += 1
        while ia < len(A):
            ans.operators += [copy.deepcopy(A[ia])]
            ia += 1
        while ib < len(B):
            ans.operators += [copy.deepcopy(B[ib])]
            ib += 1
        return ans

    def __iadd__(self, other):
        """
        Implement the '+=' operation.
        """
        ans = self + other
        self.operators = ans.operators
        return self

    def __mul__(self, other):
        """
        Multiplication of two operators or an operator with a number.

        Args:
            other(PolynomialOperator/int/float/complex): multiplier

        Returns:
            PolynomialOperator: self * other
        """
        ans = self.get_polynomial()
        if not isinstance(other, PolynomialOperator):
            ans.operators = [[copy.deepcopy(mono[0]), mono[1] * other] for mono in self.operators]
            return ans
        A = self.operators
        B = other.operators
        for mono_A in A:
            for mono_B in B:
                ans += self.get_polynomial(mono_A[0] + mono_B[0], mono_A[1] * mono_B[1])
        return ans

    def __imul__(self, other):
        """
        Implement the '*=' operation.
        """
        ans = self * other
        self.operators = ans.operators
        return self

    def __rmul__(self, other):
        """
        Args:
            other(PolynomialOperator/int/float/complex): multiplier

        Returns:
            PolynomialOperator: other * self
        """
        if not isinstance(other, PolynomialOperator):
            return self * other
        return other * self

    def __sub__(self, other):
        """
        Substraction of two operators.

        Args:
            other(PolynomialOperator): Operator to be substracted

        Returns:
            PolynomialOperator: self - other
        """
        return self + other * (-1)

    def __isub__(self, other):
        """
        Implement the '-=' operation.
        """
        self += other * (-1)
        return self

    def __truediv__(self, other):
        """
        Division of an operator with a number.

        Args:
            other(int/float/complex): divisor

        Returns:
            PolynomialOperator: self / other
        """
        return self * (1. / other)

    def __itruediv__(self, other):
        """
        Implement the '/=' operation.
        """
        self *= (1. / other)
        return self

    def __eq__(self, other) -> bool:
        """
        Judge whether two opperator polynomials are the same.
        Error within 10^(-6) can be ignored

        Args:
            other(PolynomialOperator): Operator to be judged

        Returns:
            bool: whether two opperator polynomials are the same
        """
        if not isinstance(other, PolynomialOperator):
            return False
        delta = self - other
        for mono in delta.operators:
            if (abs(mono[1]) > 1e-6):
                return False
        return True

    def parse(self):
        """
        Simply parse the list of the operators to string.
        """
        return f"{self.operators}"

    def __repr__(self) -> str:
        """
        Parse the list format to string format.
        """
        if self.operators == []:
            return '= 0 '
        ans = ''
        for mono in self.operators:
            ans += '+ (' + str(mono[1]) + ') '
            if mono[0] != []:
                ans += '* '
                for var in mono[0]:
                    ans += self.parse_single(var)
        ans = '=' + ans[1:]
        return ans



from typing import Union

from .data_def import TrainConfig
from train.learner import Learner

from QuICT.core import *
from QuICT.core.gate import CompositeGate
from QuICT.qcda.utility import OutputAligner


class RlMapping:
    def __init__(self, layout: Layout, inference_model_path: str = "./model") -> None:
        self._config = TrainConfig(
            topo=layout, inference=True, inference_model_dir=inference_model_path
        )
        self._learner = Learner(config=self._config)

    @OutputAligner()
    def execute(
        self, circuit_like: Union[Circuit, CompositeGate]
    ) -> Union[Circuit, CompositeGate]:
        assert (
            circuit_like.width() == self._config.topo.qubit_number
        ), "Circuit and Layout must have the same qubit number!"
        cutoff = circuit_like.width() * len(circuit_like.gates)
        mapped, remained = self._learner.map_all(
            circ=circuit_like,
            layout=self._config.topo,
            policy_net=self._learner._policy_net,
            policy_net_device=self._config.device,
            cutoff=cutoff,
        )
        if len(remained.gates) > 0:
            raise Exception("Failed to map this circuit")
        return mapped



from enum import Enum
import numpy as np

from QuICT.ops.linalg.cpu_calculator import dot


def is_kraus_ops(kraus: list) -> bool:
    row, col = kraus[0].shape
    n = int(np.log2(row))
    if not row == col or not row == 2 ** n:
        return False

    for kmat in kraus:
        if not kmat.shape == (row, col):
            return False

    kk = sum(dot(np.transpose(k).conjugate(), (k)) for k in kraus)
    if not np.allclose(kk, np.identity(row, dtype=kraus[0].dtype), rtol=1e-4):
        return False

    return True


class NoiseChannel(Enum):
    unitary = "Unitary Channel"
    pauil = "Pauli Channel"
    depolarizing = "Depolarizing Channel"
    damping = "Damping Channel"
    readout = "Readout Channel"



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/08/14 18:30
# @Author  : Xiaoquan Xu
# @File    : qubit_operator.py

"""
A Qubit operator is a polynomial of Pauli matrices {X, Y, Z} = {sigma_1, sigma_2, sigma_3},
which is a useful representation for circuits by second quantization.
"""

from QuICT.chemistry.operator.polynomial_operator import PolynomialOperator


class QubitOperator(PolynomialOperator):
    """
    A Qubit operator is a polynomial of Pauli matrices {X, Y, Z} = {sigma_1, sigma_2, sigma_3},
    which is a useful representation for circuits by second quantization.

    In this class, the operator could be represented as below.
    For example, list
    [[[(i, 1), (j, 1), (k, 3), (l, 2)], 1.2], [[(i, 1), (j, 3), (s,2)], -1.2], ...]
    stands for '1.2 Xi Xj Zk Yl - 1.2 Xi Zj Ys + ...',

    In the following descriptions, the above list is called list format,
    while the above string is called string format.
    """
    def __init__(self, monomial=None, coefficient=1.):
        """
        Create a monomial of qubit operators with the two given formats.

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        """
        super().__init__(monomial, coefficient)
        if self.operators == []:
            return
        variables = self.operators[0][0]
        l = len(variables)

        # The second parameter(kind) in fermion operator should be {1,2,3}.
        if any([var[1] not in [1, 2, 3] for var in variables]):
            raise ValueError

        # The variables in a monomial should be in ascending order.
        # Commutation relation for operators on different targets.
        for i in range(l - 1, 0, -1):
            fl = False
            for j in range(i):
                if variables[j][0] > variables[j + 1][0]:
                    variables[j], variables[j + 1] = variables[j + 1], variables[j]
                    fl = True
            if not fl:
                break

        # Commutation relation for operators on identical targets.
        operators = []
        for i in range(l):
            if i == 0 or variables[i][0] != variables[i - 1][0]:
                cur = variables[i][1]
                j = i + 1
                while j < l and variables[j][0] == variables[i][0]:
                    if cur == 0:
                        cur = variables[j][1]
                    elif cur == variables[j][1]:
                        cur = 0
                    else:
                        coefficient *= complex(0, (-1) ** ((cur - variables[j][1] + 3) % 3))
                        cur = 6 - cur - variables[j][1]
                    j += 1
                if cur != 0:
                    operators += [(variables[i][0], cur)]
        self.operators = [[operators, coefficient]]

    @classmethod
    def get_polynomial(cls, monomial=None, coefficient=1.):
        '''
        Construct an instance of the same class as 'self'.

        Args:
            monomial(list/str): Operator monomial in list/string format
            coefficient(int/float/complex): Coefficient of the monomial
        '''
        return QubitOperator(monomial, coefficient)

    @classmethod
    def analyze_single(cls, single_operator):
        """
        Transform a string format of a single operator to a tuple.
        For example,
        'X12' -> (12,1); 'Y0' -> (0,2); 'Z5' -> (5,3)

        Args:
            single_operator(str): string format

        Returns:
            tuple: the corresponding tuple in list format
        """
        if single_operator[0] == 'X':
            return (int(single_operator[1:]), 1)
        elif single_operator[0] == 'Y':
            return (int(single_operator[1:]), 2)
        elif single_operator[0] == 'Z':
            return (int(single_operator[1:]), 3)
        else:
            raise Exception("The string format is not recognized: " + single_operator)

    @classmethod
    def parse_single(cls, single_operator):
        """
        Transform a tuple format of a single operator to a string.
        For example,
        (12,1) -> 'X12 '; (0,2) -> 'Y0 '; (5,3) -> 'Z5 '

        Args:
            single_operator(tuple): list format

        Returns:
            string: the corresponding string format
        """
        if single_operator[1] == 1:
            return 'X' + str(single_operator[0]) + ' '
        elif single_operator[1] == 2:
            return 'Y' + str(single_operator[0]) + ' '
        elif single_operator[1] == 3:
            return 'Z' + str(single_operator[0]) + ' '



# !/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/4/27 2:02 下午
# @Author  : Han Yu
# @File    : _simulator
import numpy as np
from typing import Union

from QuICT.core import Circuit
from QuICT.core.noise import NoiseModel
from QuICT.simulation.state_vector import CircuitSimulator
from QuICT.simulation.unitary import UnitarySimulator
from QuICT.simulation.density_matrix import DensityMatrixSimulation
from QuICT.simulation.utils import Result, options_validation
from QuICT.tools.exception.core import ValueError
from QuICT.tools.exception.simulation import SimulatorOptionsUnmatchedError


class Simulator:
    """ The high-level simulation class, including all QuICT simulator mode.

    Args:
        device (str): The device of the simulator. One of [CPU, GPU]
        backend (str): The backend for the simulator. One of [unitary, state_vector, density_matrix]
        shots (int): The running times; must be a positive integer, default to 1.
        precision (str): The precision of simulator, one of [single, double], default to double.
        circuit_record (bool): whether record circuit's qasm in output, default to False.
        amplitude_record (bool): whether record the amplitude of qubits, default to False.
        **options (dict): other optional parameters for the simulator.
    """

    __DEVICE = ["CPU", "GPU"]
    __BACKEND = ["unitary", "state_vector", "density_matrix"]
    __PRECISION = ["single", "double"]

    def __init__(
        self,
        device: str = "CPU",
        backend: str = "state_vector",
        precision: str = "double",
        circuit_record: bool = False,
        amplitude_record: bool = False,
        output_path: str = None,
        **options
    ):
        assert device in Simulator.__DEVICE, ValueError("Simulator.device", "[CPU, GPU]", device)
        self._device = device
        assert backend in Simulator.__BACKEND, \
            ValueError("Simulator.backend", "[unitary, state_vector, density_matrix]", backend)
        self._backend = backend
        assert precision in Simulator.__PRECISION, ValueError("Simulator.precision", "[single, double]", precision)
        self._precision = precision

        if options_validation(options=options, device=self._device, backend=self._backend):
            self._options = options
        else:
            raise SimulatorOptionsUnmatchedError(
                f"Unmatched options arguments depending on {self._device} and {self._backend}."
            )

        # Result's arguments
        self._result_recorder = Result(
            device, backend, precision, circuit_record, amplitude_record, self._options, output_path
        )

    def _load_simulator(self):
        """ Initial simulator. """
        if self._device == "GPU":
            from QuICT.simulation.state_vector import ConstantStateVectorSimulator

        if self._backend == "state_vector":
            simulator = ConstantStateVectorSimulator(precision=self._precision, **self._options) \
                if self._device == "GPU" else CircuitSimulator(precision=self._precision)
        else:
            if self._backend == "unitary":
                simulator = UnitarySimulator(device=self._device, precision=self._precision)
            else:
                simulator = DensityMatrixSimulation(device=self._device, precision=self._precision)

        return simulator

    def run(
        self,
        circuit: Union[Circuit, np.ndarray],
        shots: int = 1,
        state_vector: np.ndarray = None,
        density_matrix: np.ndarray = None,
        noise_model: NoiseModel = None,
        use_previous: bool = False
    ):
        """ start simulator with given circuit

        Args:
            circuit Union[Circuit, np.ndarray]: The quantum circuits or unitary matrix.
            state_vector (ndarray): The initial state vector.
            density_matrix (ndarray): The initial density matrix.
            noise_model (NoiseModel, optional): The NoiseModel only for density_matrix simulator. Defaults to None.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.

        Yields:
            [dict]: The Result Dict.
        """
        if isinstance(circuit, np.ndarray) and self._backend != "unitary":
            raise SimulatorOptionsUnmatchedError(
                f"The unitary matrix input only allows in the unitary backend, not {self._backend}."
            )

        if (density_matrix is not None or noise_model is not None) and self._backend != "density_matrix":
            raise SimulatorOptionsUnmatchedError(
                "The density matrix and noise model input only allows in the density_matrix backend, " +
                f"not {self._backend}."
            )

        if state_vector is not None and self._backend == "density_matrix":
            raise SimulatorOptionsUnmatchedError(
                "The state vector input is not allowed in the density matrix backend."
            )

        if isinstance(circuit, Circuit):
            self._result_recorder.record_circuit(circuit)

        simulator = self._load_simulator()
        if self._backend == "density_matrix":
            amplitude = simulator.run(circuit, density_matrix, noise_model, use_previous)
        else:
            amplitude = simulator.run(circuit, state_vector, use_previous)

        self._result_recorder.record_amplitude(amplitude)

        sample_result = simulator.sample(shots)
        self._result_recorder.record_sample(sample_result)

        return self._result_recorder.__dict__()



from enum import Enum


class GateType(Enum):
    h = "H gate"
    hy = "Self-inverse gate"
    s = "S gate"
    sdg = "The conjugate transpose of Phase gate"
    x = "Pauli-X gate"
    y = "Pauli-Y gate"
    z = "Pauli-Z gate"
    sx = "sqrt(X) gate"
    sy = "sqrt(Y) gate"
    sw = "sqrt(W) gate"
    id = "Identity gate"
    u1 = "U1 gate"
    u2 = "U2 gate"
    u3 = "U3 gate"
    rx = "Rx gate"
    ry = "Ry gate"
    rz = "Rz gate"
    ri = "Ri gate"
    t = "T gate"
    tdg = "The conjugate transpose of T gate"
    phase = "Phase gate"
    gphase = "Global Phase gate"
    cz = "controlled-Z gate"
    cx = "controlled-X gate"
    cy = "controlled-Y gate"
    ch = "controlled-Hadamard gate"
    crz = "controlled-Rz gate"
    cu1 = "controlled-U1 gate"
    cu3 = "controlled-U3 gate"
    fsim = "fSim gate"
    rxx = "Rxx gate"
    ryy = "Ryy gate"
    rzz = "Rzz gate"
    rzx = "Rzx gate"
    swap = "Swap gate"
    cswap = "cswap gate"
    iswap = "iswap gate"
    iswapdg = "The conjugate transpose of iswap gate"
    sqiswap = "square root of iswap gate"
    ccx = "Toffoli gate"
    ccz = "Multi-Control Z Gate"
    ccrz = "CCRz gate"

    # Special gate below
    measure = "Measure gate"
    reset = "Reset gate"
    barrier = "Barrier gate"
    unitary = "Unitary gate"

    # no qasm represent below
    perm = "Permutation gate"
    perm_fx = "Perm-Fx gate"

    # Composite gate
    qft = "QFT gate"
    iqft = "IQFT gate"


class MatrixType(Enum):
    """ Different Type of quantum gates' matrix

    normal: based type of matrix
        1-bits: [[a,b], [c,d]]
        2-bits(control): [[1, 0, 0, 0],
                          [0, 1, 0, 0],
                          [0, 0, a, b],
                          [0, 0, c, d]]
    diagonal: diagonal matrix
        1-bits: [a, 0], [0, b]
        2-bits(control): [[1, 0, 0, 0],
                          [0, 1, 0, 0],
                          [0, 0, a, 0],
                          [0, 0, 0, b]]
        2-bits(targets): [[a, 0, 0, 0],
                          [0, b, 0, 0],
                          [0, 0, c, 0],
                          [0, 0, 0, d]]
        3-bits (control, target); [1, 1, 1, 1, 1, 1, a, b] -- diagonal values
    control: control diagonal matrix
        1-bits: [[1, 0], [0, a]]
        2-bits(control): [[1, 0, 0, 0],
                          [0, 1, 0, 0],
                          [0, 0, 1, 0],
                          [0, 0, 0, a]]
    swap: swap quantum gates' matrix
        1-bit [x]: [[0, 1], [1, 0]]
        2-bit [swap]: [[1, 0, 0, 0],
                       [0, 0, 1, 0],
                       [0, 1, 0, 0],
                       [0, 0, 0, 1]]
        3-bit [cswap]: [[ID(4)],
                        ,   [1, 0, 0, 0],
                            [0, 0, 1, 0],
                            [0, 1, 0, 0],
                            [0, 0, 0, 1]]
    reverse; reverse matrix
        1-bit: [0, a], [b, 0]
        2-bit: [[1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 0, a],
                [0, 0, b, 0]]
        3-bit: [ID(4)]
                    [1, 0, 0, 0]
                    [0, 1, 0, 0]
                    [0, 0, 0, a]
                    [0, 0, b, 0]
    special: no matrix [Measure, Reset, Barrier, Perm]
    diag_diag: 2-qubits diagonal matrix
        2-bits [Rzz]:  [[a, 0, 0, 0],
                        [0, b, 0, 0],
                        [0, 0, c, 0],
                        [0, 0, 0, d]]
    ctrl_normal: control-normal mixed quantum gate's matrix
        2-bits [FSim]: [[1, 0, 0, 0],
                        [0, a, b, 0],
                        [0, c, d, 0],
                        [0, 0, 0, A]]
    normal-normal: normal-normal mixed quantum gate's matrix
        2-bits [Rxx, Ryy]: [[A, 0, 0, B],
                            [0, a, b, 0],
                            [0, c, d, 0],
                            [C, 0, 0, D]]
    diagonal-normal: diagonal-normal mixed quantum gate's matrix
        2-bits [Rzx]: [[A, B, 0, 0],
                       [C, D, 0, 0],
                       [0, 0, a, b],
                       [0, 0, c, d]]
    """

    normal = "normal matrix"
    diagonal = "diagonal matrix"
    control = "control matrix"
    swap = "swap matrix"
    reverse = "reverse matrix"
    special = "special matrix"
    diag_diag = "diagonal * diagonal"
    ctrl_normal = "control * matrix"
    normal_normal = "normal * normal"
    diag_normal = "diagonal * normal"


SPECIAL_GATE_SET = [
    GateType.measure,
    GateType.reset,
    GateType.barrier,
    GateType.unitary,
    GateType.perm,
    GateType.perm_fx,
    GateType.qft,
    GateType.iqft,
]


DIAGONAL_GATE_SET = [
    GateType.s,
    GateType.sdg,
    GateType.z,
    GateType.id,
    GateType.u1,
    GateType.rz,
    GateType.t,
    GateType.tdg,
    GateType.phase,
    GateType.gphase,
    GateType.cz,
    GateType.crz,
    GateType.cu1,
    GateType.rzz,
    GateType.ccrz,
]


SUPREMACY_GATE_SET = [GateType.sx, GateType.sy, GateType.sw]


PAULI_GATE_SET = [GateType.x, GateType.y, GateType.z, GateType.id]


CLIFFORD_GATE_SET = [
    GateType.x,
    GateType.y,
    GateType.z,
    GateType.h,
    GateType.s,
    GateType.sdg,
    GateType.cx,
]



from .backend.uniformly_gate import *
from QuICT.core.gate import GateType
from QuICT.tools.exception.core import TypeError


class UniformlyControlGate(object):
    """
    Uniformly Ry, Rz or one-qubit Unitary gate
    """
    def __init__(self, target_gate: GateType = GateType.unitary):
        """
        Args:
            target_gate(GateType): type of target gate, could be Ry, Rz or Unitary
        """
        assert target_gate in [GateType.ry, GateType.rz, GateType.unitary], \
            TypeError("UniformlyControlGate.target_gate", [GateType.ry, GateType.rz, GateType.unitary], target_gate)
        self.target_gate = target_gate

    def __call__(self, arg_list):
        """
        Args:
            arg_list(list): a list of angles for Ry and Rz or a list of 2*2 unitaries for Unitary

        Returns:
            CompositeGate: uniformly control target gate
        """
        if self.target_gate in [GateType.ry, GateType.rz]:
            return UniformlyRotation(self.target_gate).execute(arg_list)
        else:
            return UniformlyUnitary().execute(arg_list)



import base64
import hmac

from hashlib import sha256
from Crypto.Cipher import AES
from Crypto.Util.Padding import unpad, pad


class EncryptManager:
    """ The class contains encrypt/decrypt functions of AES and SHA256. """
    def __init__(self):
        self._default_encrypted_key = "QuICTOpenSource2022"

    def encryptedmsg(self, msg: str, key: bytes) -> bytes:
        """ Encrypted the given message through AES with the key.

        Args:
            msg (str): The message which will be encrypted.
            key (bytes): The key for AES mode.

        Returns:
            bytes: The encrypted message.
        """
        if isinstance(msg, str):
            msg = msg.encode()

        if isinstance(key, str):
            key = key.encode()

        aes = AES.new(pad(key, 16), mode=AES.MODE_ECB)
        aes_message = aes.encrypt(pad(msg, 16))
        encrypted_text = base64.encodebytes(aes_message)

        return encrypted_text

    def decryptedmsg(self, msg: bytes, key: str, output_byte: bool = False) -> str:
        """ Decrypted the given message through AES with the key.

        Args:
            msg (bytes): The encrypted message which will be decrypted.
            key (str): The key for AES mode.
            output_byte (bool, optional): Whether return the decrypted in byte or string. Defaults to False.

        Returns:
            Union[str, byte]: The decrypted message.
        """
        if isinstance(msg, str):
            msg = msg.encode()

        if isinstance(key, str):
            key = key.encode()

        aes = AES.new(pad(key, 16), mode=AES.MODE_ECB)
        base64_decryptedmsg = base64.decodebytes(msg)
        encrypted_msg = unpad(aes.decrypt(base64_decryptedmsg), 16)
        if not output_byte:
            encrypted_msg = str(encrypted_msg, encoding='utf-8')

        return encrypted_msg

    def encrypted_passwd(self, passwd: str) -> str:
        """ Encrypt the password through SHA256

        Args:
            passwd (str): The given password.

        Returns:
            str: the encrypted password.
        """
        key = self._default_encrypted_key.encode('utf-8')
        if isinstance(passwd, str):
            passwd = passwd.encode('utf-8')

        encrypted_passwd = hmac.new(key, passwd, digestmod=sha256).digest()
        sign = base64.b64encode(encrypted_passwd).decode()

        return sign



import sys

from QuICT.tools import Logger, LogFormat
from QuICT.simulation import Simulator
from QuICT.tools.interface import OPENQASMInterface


logger = Logger("Simulation_Local_Mode", LogFormat.full)


def simulation_start(
    circuit_path: str,
    shots: str,
    device: str,
    backend: str,
    precision: str,
    output_path: str
):
    logger.info("Start Run Simulation Job in local mode.")
    logger.debug(
        f"Job Parameters: circuit path: {circuit_path}, shots: {shots}, " +
        f"Precision: {precision}, Device: {device}, Backend: {backend}, " +
        f"output_path: {output_path}."
    )
    # Get circuit from given path
    circuit = OPENQASMInterface.load_file(circuit_path).circuit

    # Start Simulator
    simulator = Simulator(
        device=device,
        backend=backend,
        precision=precision,
        output_path=output_path,
        amplitude_record=True
    )
    simulator.run(circuit, shots=int(shots))

    logger.info(f"Simulation Job finished, store the result in {output_path}.")


if __name__ == "__main__":
    raw_args = sys.argv[1:]
    dict_args = dict([arg.split('=', maxsplit=1) for arg in raw_args])
    logger.info(f"{dict_args}")

    simulation_start(
        **dict_args
    )



from QuICT.tools.exception import QuICTException


class KrausError(QuICTException):
    """ Not Kraus Matrix Error. """
    def __init__(self, msg: str = None):
        super().__init__(1012, msg)


class NoiseApplyError(QuICTException):
    """ Noise Apply Error. """
    def __init__(self, msg: str = None):
        super().__init__(1013, msg)


class PauliNoiseUnmatchedError(QuICTException):
    """ Pauli Noise operators and number qubits unmatched error. """
    def __init__(self, msg: str = None):
        super().__init__(1014, msg)


class DamplingNoiseMixedProbExceedError(QuICTException):
    """ Dampling Noise' amplitude prob and phase prob sum exceed 1. """
    def __init__(self, msg: str = None):
        super().__init__(1015, msg)



import copy
import math
import numpy as np
from typing import Union, List

from QuICT.ops.linalg.cpu_calculator import dot, tensor
from QuICT.tools.exception.core import ValueError, TypeError
from .utils import NoiseChannel


class ReadoutError:
    """ The Readout error class

    Example:
        p(n|m) describe the probability of getting the noise outcome n with the truly measured result m. \n
        The ReadoutError for 1 qubit: \n
            P = [[p(0|0), p(1|0)], = [[0.8, 0.2], \n
                 [p(0|1), p(1|1)]]    [0.3, 0.7]]

        The ReadoutError for 2 qubit:
            P = [[p(00|00), p(01|00), p(10|00), p(11|00)], \n
                 [p(00|01), p(01|01), p(10|01), p(11|01)], \n
                 [p(00|10), p(01|10), p(10|10), p(11|10)], \n
                 [p(00|11), p(01|11), p(10|11), p(11|11)]] \n

    Important:
        The sum of each rows in the prob should equal to 1.

    Args:
        prob (List, np.ndarray): The probability of outcome assignment
    """
    @property
    def qubits(self) -> int:
        return self._qubits

    @property
    def prob(self) -> np.ndarray:
        return self._prob

    @property
    def type(self) -> str:
        return self._type

    def __init__(self, prob: Union[List, np.ndarray]):
        self._prob = self._probability_check(prob)
        self._qubits = int(np.log2(self._prob.shape[0]))
        self._type = NoiseChannel.readout

    def __str__(self):
        ro_str = f"{self.type.value} with {self.qubits} qubits.\n"
        for i, probs in enumerate(self.prob):
            ro_str += f"P[{i}]: {probs}\n"

        return ro_str

    def _probability_check(self, prob):
        if not isinstance(prob, (list, np.ndarray)):
            raise TypeError("ReadoutError.prob", "list or np.ndarray.", f"{type(prob)}")

        prob = np.array(prob)
        row, col = prob.shape
        if row != col:
            raise ValueError("ReadoutError.prob.shape", "row == col", f"{row} and {col}")

        n = int(np.log2(row))
        if 2 ** n != row:
            raise ValueError("ReadoutError.prob.shape", "2 ** n", f"{row}")

        for p in prob:
            if not isinstance(prob, (list, np.ndarray)):
                raise TypeError("ReadoutError.prob", "list or np.ndarray.", f"{type(prob)}")

            sum_p = [i for i in p if i >= 0 and i <= 1]
            if not math.isclose(sum(sum_p), 1, rel_tol=1e-6) or len(sum_p) != len(p):
                raise ValueError("ReadoutError.prob", "same to 1 for each row", f"{sum(sum_p)}")

        return prob

    def expand(self, extend_qubits: int, change_itself: bool = False):
        """ Expand noise with size of extend_qubits. """
        extra_identity = np.identity(2 ** extend_qubits)
        expand_prob = tensor(self.prob, extra_identity)

        if change_itself:
            self._prob = expand_prob
            self._qubits += extend_qubits
        else:
            return ReadoutError(expand_prob)

    def compose(self, other):
        """ dot(self.prob, other.prob) """
        assert isinstance(other, ReadoutError)

        left_prob, right_prob = self.prob, other.prob
        if self.qubits > other.qubits:
            right_prob = other.expand(self.qubits - other.qubits).prob
        elif self.qubits < other.qubits:
            left_prob = self.expand(other.qubits - self.qubits).prob

        compose_prob = dot(left_prob, right_prob)
        return ReadoutError(compose_prob)

    def tensor(self, other):
        """ tensor(self.prob, other.prob) """
        assert isinstance(other, ReadoutError)

        tensor_prob = tensor(self.prob, other.prob)
        return ReadoutError(tensor_prob)

    def power(self, n: int):
        """ (self.prob)^n """
        assert isinstance(n, int)

        based_prob = copy.deepcopy(self.prob)
        for _ in range(1, n):
            based_prob = dot(based_prob, self.prob)

        return ReadoutError(based_prob)

    def is_identity(self) -> bool:
        """ Whether self.prob is identity matrix. """
        id_matrix = np.identity(2 ** self._qubits)
        if np.allclose(self._prob, id_matrix, rtol=1e-6):
            return True

        return False

    def apply_to_qubits(self, measured_result: int):
        prob = np.random.random()
        related_prob_error = self._prob[measured_result]
        for idx, error_prob in enumerate(related_prob_error):
            if prob <= error_prob:
                return idx

            prob -= error_prob



from typing import Dict

from QuICT.core.gate import CompositeGate
from ._operator import Operator


class DeviceTrigger(Operator):
    """ The mapping of Quantum gates and device id. """
    def __init__(
        self,
        device_gate_mapping: Dict[int, CompositeGate]
    ):
        """
        Args:
            targets (int): The number of target qubits.
            device_gate_mapping (Dict[int, CompositeGate]): The mapping of device and related composite gates.

        Raises:
            TypeError: Error input parameters.
        """
        super().__init__(targets=1)
        self._dev_to_gate = device_gate_mapping

    def __str__(self):
        totals = ""
        for k, v in self._dev_to_gate.items():
            string = str(k) + ": "
            for gate in v.gates:
                string += gate.qasm()

            totals += string + "\n"

        return totals

    def mapping(self, dev: int) -> CompositeGate:
        return self._dev_to_gate[dev]



import sys

from QuICT.tools.cli.script.qcda import qcda_start
from QuICT.tools.cli.script.simulation import simulation_start


def start_pipeline(**kwargs):
    simulation_related = {}
    for key in ["shots", "device", "backend", "precision"]:
        simulation_related[key] = kwargs[key]
        del kwargs[key]

    # Run QCDA process first
    qcda_start(**kwargs)

    # Run Simulation with optimized circuit from QCDA process
    output_path = kwargs['output_path']
    opt_cirpath = f"{output_path}/opt_circuit.qasm"
    simulation_start(circuit_path=opt_cirpath, output_path=output_path, **simulation_related)


if __name__ == "__main__":
    raw_args = sys.argv[1:]
    dict_args = dict([arg.split('=', maxsplit=1) for arg in raw_args])

    start_pipeline(**dict_args)



from .backward_match import BackwardMatch, MatchingScenario
from .forward_match import ForwardMatch
from .matching_dag_circuit import Match, MatchingDAGCircuit, MatchingDAGNode
from .template_matching import TemplateMatching
from .template_substitution import TemplateSubstitution



import string
import random
import smtplib
from email.mime.text import MIMEText
from email.header import Header


EMAIL_HOST = "smtp.sina.com"
MAIL_USER = "quict_pin_reset"
AUTH_KEY = "ae2725b5ce956101"
SENDER_EMAIL = "quict_pin_reset@sina.com"


def generator_random_password() -> str:
    default_size = 16
    chars = string.ascii_letters + string.digits
    rand_chars = [random.choice(chars) for _ in range(default_size)]

    return "".join(rand_chars)


def send_reset_password_email(receiver: str) -> str:
    """ Send reset password emails to receiver. """
    reset_password = generator_random_password()

    message = MIMEText(
        "Hi, there: \n" +
        f"The new password for QuICT is {reset_password}, please change it after successful login.\n" +
        "Thanks for your support.",
        "plain",
        "utf-8"
    )
    message["From"] = Header(SENDER_EMAIL)
    message["Subject"] = Header("QuICT User Password Reset", "utf-8")

    try:
        smtpObj = smtplib.SMTP()                # Initial SMTP Object
        smtpObj.connect(EMAIL_HOST, 25)         # Connect SMTP Server
        smtpObj.login(MAIL_USER, AUTH_KEY)      # Validation login
        smtpObj.sendmail(SENDER_EMAIL, receiver, message.as_string())    # Send Context.

        return reset_password
    except smtplib.SMTPException as error:
        print("error:{}".format(error))



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# Modification Notice: Code revised for QuICT

"""QASM nodes."""

from .barrier import Barrier
from .binaryop import BinaryOp
from .binaryoperator import BinaryOperator
from .cnot import Cnot
from .creg import Creg
from .customunitary import CustomUnitary
from .expressionlist import ExpressionList
from .external import External
from .gate import Gate
from .gatebody import GateBody
from .id import Id
from .idlist import IdList
from .if_ import If
from .indexedid import IndexedId
from .intnode import Int
from .format import Format
from .measure import Measure
from .opaque import Opaque
from .prefix import Prefix
from .primarylist import PrimaryList
from .program import Program
from .qreg import Qreg
from .real import Real
from .reset import Reset
from .unaryoperator import UnaryOperator
from .universalunitary import UniversalUnitary
from .nodeexception import NodeException



from typing import Tuple, Union, List, Dict
from types import FunctionType

from ._operator import Operator
from QuICT.core.gate import CompositeGate, BasicGate
from QuICT.tools.exception.core import TypeError, ValueError


class Trigger(Operator):
    """
    The trigger for switch the dynamic circuit; contains the target qubits and
    related circuits with different state.
    """
    def __init__(
        self,
        targets: int,
        state_gate_mapping: Union[Dict[int, CompositeGate], List[CompositeGate], Tuple[CompositeGate], FunctionType]
    ):
        """
        Args:
            targets (int): The number of target qubits.
            state_gate_mapping: The mapping of state and related composite gates.
                (Union[Dict[int, CompositeGate], List[CompositeGate], Tuple[CompositeGate], FunctionType])

        Raises:
            TypeError: Error input parameters.
        """
        super().__init__(targets=targets)

        # Deal with state - compositegate mapping
        self._state_gate_mapping = {}
        if isinstance(state_gate_mapping, (list, tuple)):
            for idx, cgate in enumerate(state_gate_mapping):
                assert isinstance(cgate, (CompositeGate, BasicGate, type(None))), TypeError(
                    "Trigger.state_gate_mapping:list", "list/tuple<CompositeGate, BasicGate>", type(cgate)
                )
                self._state_gate_mapping[idx] = cgate
        elif isinstance(state_gate_mapping, dict):
            for key, value in state_gate_mapping.items():
                assert isinstance(key, int) and isinstance(value, (CompositeGate, BasicGate, type(None))), \
                    TypeError(
                        "Trigger.state_gate_mapping:dict",
                        "dict<int, CompositeGate/BasicGate/None>",
                        f"{type(key)}, {type(value)}")

            self._state_gate_mapping = state_gate_mapping
        elif isinstance(state_gate_mapping, FunctionType):
            self._check_function_validation(state_gate_mapping)
            self._state_gate_mapping = state_gate_mapping
        else:
            raise TypeError("Trigger.state_gate_mapping", "[list, dict, Function]", {type(state_gate_mapping)})

    def mapping(self, state: int) -> CompositeGate:
        """ Return the related composite gate with given qubits' measured state.

        Args:
            state (int): The qubits' measured state.

        Returns:
            CompositeGate: The related composite gate.
        """
        assert state >= 0 and state < 2 ** self.targets, ValueError(
            "Trigger.mapping.state", f"[0, {2**self.targets}]", state
        )

        if isinstance(self._state_gate_mapping, FunctionType):
            return self._state_gate_mapping(state)
        else:
            return self._state_gate_mapping[state]

    def _check_function_validation(self, state_gate_mapping):
        """ Validation the correctness of given state-composite mapping function. """
        for i in range(2 ** self.targets):
            if not isinstance(state_gate_mapping(i), (CompositeGate, BasicGate, type(None))):
                raise TypeError(
                    "Trigger.state_gate_mapping:callable",
                    "Return(CompositeGate/BasicGate/None)",
                    type(state_gate_mapping(i))
                )



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/8/22 2:40
# @Author  : Han Yu
# @File    : __init__.py

from .clifford_rz_optimization import CliffordRzOptimization
from .commutative_optimization import CommutativeOptimization
from .symbolic_clifford_optimization import SymbolicCliffordOptimization
from .cnot_ancilla import CnotAncilla
from .cnot_local_force import (CnotForceBfs, CnotForceDepthBfs,
                               CnotLocalForceBfs, CnotLocalForceDepthBfs,
                               CnotStoreForceBfs, CnotStoreForceDepthBfs)
from .cnot_without_ancilla import CnotWithoutAncilla
from .template_optimization import TemplateOptimization



from .. import InstructionSet
from ..transform_rule import xyx_rule

from QuICT.core.gate import *

IonQSet = InstructionSet(
    GateType.rxx,
    [GateType.rx, GateType.ry, GateType.rz]
)
IonQSet.register_one_qubit_rule(xyx_rule)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/12 6:33
# @Author  : Han Yu
# @File    : MCT_Linear_Simulation.py

from QuICT.core import *
from QuICT.core.gate import *


class MCTLinearHalfDirtyAux(object):
    def execute(self, m: int, n: int):
        """ A linear simulation for Toffoli gate

        https://arxiv.org/abs/quant-ph/9503016 Lemma 7.2

        Implement a m-bit toffoli gate in a qureg with n qubit with linear complexity.

        If n ≥ 5 and m ∈ {3, . . . , ⌈n/2⌉} then (m+1)-Toffoli gate can be simulated
        by a network consisting of 4(m - 2) toffoli gates

        Args:
            m(int): the number of control qubits of the toffoli
            n(int): the number of qubits in the qureg

        Returns:
            CompositeGate: the result of Decomposition
        """
        if m > (n // 2) + (1 if n % 2 == 1 else 0):
            raise Exception("control bit cannot above ceil(n/2)")
        if m < 1:
            raise Exception("there must be at least one control bit")

        controls = [i for i in range(m)]
        auxs = [i for i in range(m, n - 1)]
        target = n - 1

        return self.assign_qubits(n, m, controls, auxs, target)

    @staticmethod
    def assign_qubits(n: int, m: int, controls: list, auxs: list, target: int):
        """
        Args:
            n(int): the number of qubits in the qureg
            m(int): the number of control qubits of the toffoli
            controls(list): list of control qubits
            auxs(list): list of ancillas
            target(int): target qubit

        Returns:
            CompositeGate: the result of Decomposition
        """
        gates = CompositeGate()
        with gates:
            qubit_list = controls + auxs + [target]
            if m == 1:
                CX & [qubit_list[0], qubit_list[n - 1]]
            elif m == 2:
                CCX & [qubit_list[0], qubit_list[1], qubit_list[n - 1]]
            else:
                for i in range(m, 2, -1):
                    CCX & [qubit_list[i - 1], qubit_list[n - 1 - (m - i + 1)], qubit_list[n - 1 - (m - i)]]
                CCX & [qubit_list[0], qubit_list[1], qubit_list[n - m + 1]]
                for i in range(3, m + 1):
                    CCX & [qubit_list[i - 1], qubit_list[n - 1 - (m - i + 1)], qubit_list[n - 1 - (m - i)]]

                for i in range(m - 1, 2, -1):
                    CCX & [qubit_list[i - 1], qubit_list[n - 1 - (m - i + 1)], qubit_list[n - 1 - (m - i)]]
                CCX & [qubit_list[0], qubit_list[1], qubit_list[n - m + 1]]
                for i in range(3, m):
                    CCX & [qubit_list[i - 1], qubit_list[n - 1 - (m - i + 1)], qubit_list[n - 1 - (m - i)]]

        return gates


class MCTLinearOneDirtyAux(object):
    def execute(self, n):
        """ A linear simulation for Toffoli gate

        https://arxiv.org/abs/quant-ph/9503016 Corollary 7.4

        Implement an n-bit toffoli gate in a qureg with n + 2 qubits with linear complexity.

        On an n-bit circuit, an (n-2)-bit toffoli gate can be simulated by
        8(n-5) CCX gates, with 1 bit dirty ancilla.

        Args:
            n(int): the number of used qubit, which is (n + 2) for n-qubit Toffoli gates

        Returns:
            CompositeGate: the result of Decomposition
        """
        if n < 3:
            raise Exception("there must be at least one control bit")

        controls = list(range(n - 2))
        target = n - 2
        aux = n - 1     # this is a dirty ancilla

        gates = CompositeGate()
        with gates:
            n = len(controls) + 2
            if n == 5:
                CCX & [controls[0], controls[1], aux]
                CCX & [controls[2], aux, target]
                CCX & [controls[0], controls[1], aux]
                CCX & [controls[2], aux, target]
                return gates
            if n == 4:
                CCX & [controls[0], controls[1], target]
                return gates
            if n == 3:
                CX & [controls[0], target]
                return gates
            # n > 5
            m = n // 2
            MCT_half_dirty = MCTLinearHalfDirtyAux()
            gates_first = MCT_half_dirty.assign_qubits(n, m, controls[0:m], controls[m:n - 2] + [target], aux)
            gates_last = MCT_half_dirty.assign_qubits(n, n - m - 1, controls[m:n - 2] + [aux], controls[0:m], target)
            if n - m == 3:  # n == 6
                gates_first | gates
                CCX & [controls[m], aux, target]
                gates_first | gates
                CCX & [controls[m], aux, target]
            else:
                gates_first | gates
                gates_last | gates
                gates_first | gates
                gates_last | gates

        return gates



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# Modification Notice: Code revised for QuICT

"""Node for an OPENQASM id."""

from .node import Node
from .nodeexception import NodeException


class Id(Node):
    """Node for an OPENQASM id.

    The node has no children but has fields name, line, and file.
    There is a flag is_bit that is set when XXXXX to help with scoping.
    """

    def __init__(self, id, line, file):
        """Create the id node."""
        # pylint: disable=redefined-builtin
        super().__init__("id", None, None)
        self.name = id
        self.line = line
        self.file = file
        # To help with scoping rules, so we know the id is a bit,
        # this flag is set to True when the id appears in a gate declaration
        self.is_bit = False

    def to_string(self, indent):
        """Print the node with indent."""
        ind = indent * ' '
        print(ind, 'id', self.name)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        del prec  # prec ignored
        return self.name

    def latex(self, prec=15, nested_scope=None):
        """Return the correspond math mode latex string."""
        if not nested_scope:
            return "\textrm{" + self.name + "}"
        else:
            if self.name not in nested_scope[-1]:
                raise NodeException("Expected local parameter name: ",
                                    "name=%s, " % self.name,
                                    "line=%s, " % self.line,
                                    "file=%s" % self.file)

            return nested_scope[-1][self.name].latex(prec,
                                                     nested_scope[0:-1])

    def sym(self, nested_scope=None):
        """Return the correspond symbolic number."""
        if not nested_scope or self.name not in nested_scope[-1]:
            raise NodeException("Expected local parameter name: ",
                                "name=%s, line=%s, file=%s" % (
                                    self.name, self.line, self.file))

        return nested_scope[-1][self.name].sym(nested_scope[0:-1])

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        if not nested_scope or self.name not in nested_scope[-1]:
            raise NodeException("Expected local parameter name: ",
                                "name=%s, line=%s, file=%s" % (
                                    self.name, self.line, self.file))

        return nested_scope[-1][self.name].real(nested_scope[0:-1])



import time
from collections import deque
from functools import cached_property
from typing import List

from QuICT.qcda.optimization.commutative_optimization import \
    CommutativeOptimization
from QuICT.qcda.utility import OutputAligner

from .dag import DAG
from .symbolic_phase import SymbolicPhase
from .template import *


class CliffordRzOptimization(object):
    """
    Heuristic optimization of circuits in Clifford + Rz.

    [1] Nam, Yunseong, et al. "Automated optimization of large quantum
    circuits with continuous parameters." npj Quantum Information 4.1
    (2018): 1-12.
    """

    _optimize_sub_method = {
        1: "reduce_hadamard_gates",
        2: "cancel_single_qubit_gates",
        3: "cancel_two_qubit_gates",
        4: "merge_rotations",
        5: "float_rotations",
    }
    _optimize_routine = {
        'heavy': [1, 3, 2, 3, 1, 2, 4, 5],
        'light': [1, 3, 2, 3, 1, 2, 4, 3, 2],
    }

    def parameterize_all(self, gates: DAG):
        """
        Convert all applicable Rz gates into T/Tdg/S/Sdg/Z in the circuit.

        Args:
            gates(DAG): DAG of the circuit
        """
        for node in gates.topological_sort():
            if node.gate_type in [GateType.s, GateType.t, GateType.sdg, GateType.tdg, GateType.z]:
                gate_, phase_ = CommutativeOptimization.parameterize(node.get_gate())
                node.gate_type = gate_.type
                node.params = gate_.pargs
                gates.global_phase += phase_

        if not isinstance(gates.global_phase, SymbolicPhase):
            gates.global_phase %= 2 * np.pi
        else:
            gates.global_phase = np.mod(gates.global_phase, 2 * np.pi)

    @cached_property
    def hadamard_templates(self):
        return generate_hadamard_gate_templates()

    @cached_property
    def single_qubit_gate_templates(self):
        return generate_single_qubit_gate_templates()

    @cached_property
    def cnot_targ_template(self):
        return generate_cnot_targ_templates()

    @cached_property
    def cnot_ctrl_template(self):
        return generate_cnot_ctrl_templates()

    @cached_property
    def gate_preserving_rewrite_template(self):
        return generate_gate_preserving_rewrite_template()

    @cached_property
    def gate_reducing_rewrite_template(self):
        return generate_gate_reducing_rewrite_template()

    def deparameterize_all(self, gates: DAG):
        """
        Convert all applicable T/Tdg/S/Sdg/Z into Rz in the circuit.

        Args:
            gates(DAG): DAG of the circuit
        """
        for node in gates.topological_sort():
            if node.gate_type == GateType.rz and not isinstance(node.params[0], SymbolicPhase):
                compo_gate_, phase_ = CommutativeOptimization.deparameterize(node.get_gate())
                if len(compo_gate_.gates) > 1:
                    continue
                node.gate_type = compo_gate_[0].type
                node.params = compo_gate_[0].pargs
                gates.global_phase += phase_

        if not isinstance(gates.global_phase, SymbolicPhase):
            gates.global_phase %= 2 * np.pi
        else:
            gates.global_phase = np.mod(gates.global_phase, 2 * np.pi)

    def reduce_hadamard_gates(self, gates: DAG):
        """
        Reduce hadamard gates in Clifford+Rz circuits by template matching.
        Adjacent X gates will also be cancelled out.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of H gates reduced.
        """
        cnt = 0
        self.deparameterize_all(gates)
        # enumerate templates and replace every occurrence
        for template in self.hadamard_templates:
            cnt += template.replace_all(gates) * template.weight
        self.parameterize_all(gates)
        return cnt

    def cancel_single_qubit_gates(self, gates: DAG):
        """
        Merge Rz gates in Clifford+Rz circuit.
        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        cnt = 0
        for node in list(gates.topological_sort()):
            # enumerate every single qubit gate
            if node.gate_type != GateType.rz or node.flag == node.FLAG_ERASED:
                continue
            # erase the gate if degree = 0
            if np.isclose(float(node.params[0]), 0):
                node.erase()
                cnt += 1
                continue

            # try cancelling while commuting the gate with templates
            # (c_node, c_qubit): the position right before template matching
            c_node, c_qubit = node, 0
            while True:
                # (n_node, n_qubit): the start of template matching
                n_node, n_qubit = c_node.successors[c_qubit]
                # stop if reaching the end of circuit
                if not n_node.gate_type:
                    break

                # if n_node is another rz, merge and erase the original node
                if n_node.gate_type == node.gate_type:
                    n_node.params[0] = n_node.params[0] + node.params[0]
                    node.erase()
                    cnt += 1
                    break

                # template matching
                mapping = None
                for template in self.single_qubit_gate_templates:
                    mapping = mapping or template.compare(c_node.successors[c_qubit])
                    if mapping:
                        # found a sub-circuit that commutes
                        # set (c_node, c_qubit) to be the last position of this sub-circuit
                        c_node, c_qubit = template.template.end_nodes[template.anchor].predecessors[0]
                        c_node = mapping[id(c_node)]
                        break
                # found no templates. commuting fails
                if not mapping:
                    break
        return cnt

    class ReachableSet:
        """
        Data structure that manages the reachable relations in a DAG.

        For a DAG of over 10000 nodes, storing reachable relations between all nodes
        consumes excessive memories. ReachableSet will automatically manage and
        store recently used relations.
        """
        def __init__(self, succ_node=None, max_size=1000000):
            """
            Args:
                  succ_node(List[Tuple[DAG.Node, int]]): right boundary
                  max_size(int): max number of entries stored in this set.
            """
            self.succ_node = succ_node
            self.size = 0
            self.reachable = {}
            self.max_size = max_size

        def query(self, node, qubit_, other):
            """
            Ask whether `(node, qubit_)` can reach the node `other`.

            Args:
                node(DAG.Node): the starting node
                qubit_(int): the starting wire
                other(DAG.Node): the target node

            Returns:
                bool: whether reachable
            """
            key = (id(node), qubit_)
            if key not in self.reachable:
                new_set = DAG.get_reachable_set(node, qubit_, self.succ_node)
                if len(new_set) + self.size > self.max_size:
                    self.reachable.clear()
                self.reachable[key] = new_set
                self.size += len(new_set)

            return id(other) in self.reachable[key]

    def cancel_two_qubit_gates(self, gates: DAG):
        """
        Merge CX gate in Clifford+Rz circuit.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        cnt = 0
        reachable = self.ReachableSet()
        for node in list(gates.topological_sort()):
            # enumerate every cnot gate
            if node.flag == DAG.Node.FLAG_ERASED or node.gate_type != GateType.cx:
                continue

            c_ctrl_node, c_ctrl_qubit = node, 0
            c_targ_node, c_targ_qubit = node, 1
            while True:
                n_ctrl_node, n_ctrl_qubit = c_ctrl_node.successors[c_ctrl_qubit]
                n_targ_node, n_targ_qubit = c_targ_node.successors[c_targ_qubit]
                # remove two adjacent cnot gates
                if id(n_ctrl_node) == id(n_targ_node) and n_ctrl_node.gate_type == GateType.cx and \
                        n_ctrl_qubit == 0 and n_targ_qubit == 1:
                    n_ctrl_node.erase()
                    node.erase()
                    cnt += 2
                    break

                # try commuting node with templates anchored in control qubit
                mapping = None
                for template in self.cnot_ctrl_template:
                    mapping = template.compare(c_ctrl_node.successors[c_ctrl_qubit])
                    # After we find a template U after current cnot, there are two cases:
                    # Case 1:        .___.         .___.
                    #         --O----| U |--     --| U |---O--
                    #           |    |___|     =   |___|   |
                    #         --X-----------     ----------X--
                    # Case 2:        .___.         .___.
                    #         --O----|   |--     --|   |---O--
                    #           |    | U |    !=   | U |   |
                    #         --X----|___|--     --|___|---X--
                    # The template only guarantees control node 'O' can be moved across U, so case 2 is invalid.
                    # By further checking whether 'X' can reach any nodes in U, we can exclude case 2.

                    if mapping and all([not reachable.query(c_targ_node, c_targ_qubit, o) for o in mapping.values()]):
                        c_ctrl_node, c_ctrl_qubit = template.template.end_nodes[template.anchor].predecessors[0]
                        c_ctrl_node = mapping[id(c_ctrl_node)]
                        break
                    else:
                        mapping = None
                if mapping:
                    continue

                # try commuting node with templates anchored in target qubit
                for template in self.cnot_targ_template:
                    mapping = template.compare(c_targ_node.successors[c_targ_qubit])
                    # if control node can reach any node in the template, it will block commuting.
                    if mapping and all([not reachable.query(c_ctrl_node, c_ctrl_qubit, o) for o in mapping.values()]):
                        c_targ_node, c_targ_qubit = template.template.end_nodes[template.anchor].predecessors[0]
                        c_targ_node = mapping[id(c_targ_node)]
                        break
                    else:
                        mapping = None
                if not mapping:
                    break
        return cnt

    def _traverse_cnot_rz_circuit(self, anchor, flag_ofs=1):
        flag_visited = flag_ofs + 1
        flag_term = flag_ofs + 2

        anchors = {anchor.qubit_loc[0]: anchor, anchor.qubit_loc[1]: anchor}
        anchor_queue = deque(anchors.keys())

        term_node = {'predecessors': {}, 'successors': {}}
        while len(anchor_queue) > 0:
            anchor_qubit = anchor_queue.popleft()
            for neighbors in ['predecessors', 'successors']:
                c_node = anchors[anchor_qubit]
                c_node.qubit_flag[c_node.qubit_id[anchor_qubit]] = flag_visited
                while True:
                    p_node, p_qubit = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]
                    if p_node.gate_type is None or p_node.gate_type not in [GateType.cx, GateType.x, GateType.rz]:
                        term_node[neighbors][anchor_qubit] = p_node
                        p_node.qubit_flag = [flag_term] * p_node.size
                        break
                    p_node.qubit_flag[p_qubit] = flag_visited

                    if p_node.gate_type == GateType.cx:
                        o_qubit = p_node.qubit_loc[p_qubit ^ 1]
                        if o_qubit not in anchors:
                            anchors[o_qubit] = p_node
                            anchor_queue.append(o_qubit)
                    c_node = p_node

        for neighbors in ['predecessors', 'successors']:
            prune_queue = deque()
            for anchor_qubit, anchor_node in anchors.items():
                c_node = anchors[anchor_qubit]

                while True:
                    p_node, p_qubit = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]
                    if p_node.qubit_flag[p_qubit] == flag_term:
                        if id(p_node) != id(term_node[neighbors][anchor_qubit]):
                            prune_queue.append((p_node, p_qubit))
                        break

                    if p_node.gate_type == GateType.cx:
                        o_qubit = p_qubit ^ 1
                        if p_node.qubit_flag[o_qubit] != flag_visited:
                            if o_qubit == 1:  # target out of bound, skip
                                pass
                            else:  # control out of bound, terminate
                                p_node.qubit_flag[p_qubit] = flag_term
                                prune_queue.append((p_node, p_qubit))
                                break
                    c_node = p_node

            while len(prune_queue) > 0:
                c_node, c_qubit = prune_queue.popleft()
                while True:
                    p_node, p_qubit = getattr(c_node, neighbors)[c_qubit]
                    if id(p_node) == id(term_node[neighbors][p_node.qubit_loc[p_qubit]]):
                        break
                    p_node.qubit_flag[p_qubit] = flag_term
                    if p_node.gate_type is not None and p_node.gate_type == GateType.cx:
                        o_qubit = p_qubit ^ 1
                        if p_node.qubit_flag[o_qubit] == flag_visited:
                            if o_qubit == 0:
                                pass
                            else:
                                p_node.qubit_flag[o_qubit] = flag_term
                                prune_queue.append((p_node, o_qubit))
                    c_node, c_qubit = p_node, p_qubit

        left_bound = {}
        anchors = {anchor.qubit_loc[0]: anchor, anchor.qubit_loc[1]: anchor}
        anchor_queue = deque(anchors.keys())
        while len(anchor_queue) > 0:
            anchor_qubit = anchor_queue.popleft()
            for neighbors in ['predecessors', 'successors']:
                c_node = anchors[anchor_qubit]
                while True:
                    p_node, p_qubit = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]
                    if p_node.qubit_flag[p_qubit] == flag_term:
                        if neighbors == 'predecessors':
                            left_bound[anchor_qubit] = (p_node, p_qubit)
                        break
                    if p_node.gate_type == GateType.cx and all([f == flag_visited for f in p_node.qubit_flag]):
                        o_qubit = p_node.qubit_loc[p_qubit ^ 1]
                        if o_qubit not in anchors:
                            anchors[o_qubit] = p_node
                            anchor_queue.append(o_qubit)
                    c_node = p_node

        edge_count = {}
        queue = deque()
        for node_, qubit_ in left_bound.values():
            s_node, s_qubit = node_.successors[qubit_]
            if s_node.qubit_flag[s_qubit] != flag_term:
                if id(s_node) not in edge_count:
                    edge_count[id(s_node)] = sum([f == flag_visited for f in s_node.qubit_flag])
                edge_count[id(s_node)] -= 1
                if edge_count[id(s_node)] == 0:
                    queue.append(s_node)

        node_list = []
        while len(queue) > 0:
            cur = queue.popleft()
            if cur.gate_type is not None and all([f == flag_visited for f in cur.qubit_flag]):
                cur.flag = cur.FLAG_VISITED
                node_list.append(cur)
            for c_qubit in range(cur.size):
                if cur.qubit_flag[c_qubit] != flag_visited:
                    continue
                s_node, s_qubit = cur.successors[c_qubit]
                if s_node is None or s_node.qubit_flag[s_qubit] == flag_term:
                    continue
                if id(s_node) not in edge_count:
                    edge_count[id(s_node)] = sum([f == flag_visited for f in s_node.qubit_flag])
                edge_count[id(s_node)] -= 1
                if edge_count[id(s_node)] == 0:
                    queue.append(s_node)
        return node_list

    def _parse_cnot_rz_circuit(self, node_list):
        phases = {}
        first_rz = {}
        cur_phases = {}

        for node_ in node_list:
            for qubit_ in node_.qubit_loc:
                if qubit_ not in cur_phases:
                    cur_phases[qubit_] = 1 << (qubit_ + 1)

            if node_.gate_type == GateType.cx:
                cur_phases[node_.qubit_loc[1]] = cur_phases[node_.qubit_loc[1]] ^ cur_phases[node_.qubit_loc[0]]
            elif node_.gate_type == GateType.x:
                cur_phases[node_.qubit_loc[0]] = cur_phases[node_.qubit_loc[0]] ^ 1
            elif node_.gate_type == GateType.rz:
                sign = -1 if cur_phases[node_.qubit_loc[0]] & 1 else 1
                mono = (cur_phases[node_.qubit_loc[0]] >> 1)
                phases[mono] = sign * node_.params[0] + (phases[mono] if mono in phases else 0)
                if mono not in first_rz:
                    first_rz[mono] = (node_, sign)
                else:
                    node_.flag = node_.FLAG_TO_ERASE

        for phase_, pack_ in first_rz.items():
            node_, sign = pack_
            if np.isclose(float(phases[phase_]), 0):
                node_.erase()
            else:
                node_.params = [sign * phases[phase_]]

        cnt = 0
        for node_ in node_list:
            if node_.gate_type != GateType.rz or node_.flag == node_.FLAG_ERASED:
                continue
            if node_.flag == node_.FLAG_TO_ERASE:
                cnt += 1
                node_.erase()
        return cnt

    def merge_rotations(self, gates: DAG):
        """
        Merge Rz gates using phase polynomials.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        gates.reset_flag()
        gates.set_qubit_loc()
        cnt = 0
        for idx, anchor_ in enumerate(list(gates.topological_sort())):
            if anchor_.gate_type != GateType.cx or anchor_.flag == anchor_.FLAG_VISITED:
                continue

            node_list = self._traverse_cnot_rz_circuit(anchor_, idx * 2)
            cnt += self._parse_cnot_rz_circuit(node_list)

        return cnt

    def _enumerate_cnot_rz_circuit(self, gates: DAG):
        """
        Iterate over CNOT+Rz sub circuit in Clifford+Rz circuit.
        Isolated Rz gates will be ignored. Each sub circuit is
        described by a tuple (left boundary, right boundary, node count). The boundary is described
        by a list whose i-th element is the tuple (node, wire) on the i-th qubit.

        Args:
            gates(DAG): DAG of the circuit

        Returns:
            Iterator[Tuple[List[DAG.Node, int]], Tuple[List[DAG.Node, int]], int]: CNOT+Rz Sub circuit
        """

        # reset qubit location of nodes in case it is corrupted sub circuit replacement
        gates.set_qubit_loc()
        # reset node flags to FLAG_DEFAULT
        gates.reset_flag()
        for node in gates.topological_sort():
            # enumerate over every unvisited CNOT gate
            if node.flag == node.FLAG_VISITED or node.gate_type != GateType.cx:
                continue

            # set of terminate node's id
            term_node_set = set()
            # encountered CNOT gates during this iteration
            # id -> Tuple[node, bitmask], bitmask(i) represents if node's i-th qubit is reachable
            visited_cnot = {}

            # STEP 1: for each anchored qubit, traverse to the left and right boundary
            anchors = {node.qubit_loc[0]: node, node.qubit_loc[1]: node}
            anchor_queue = deque(anchors.keys())
            while len(anchor_queue) > 0:
                anchor_qubit = anchor_queue.popleft()
                for neighbors in ['predecessors', 'successors']:
                    c_node = anchors[anchor_qubit]
                    while True:
                        p_node, p_qubit = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]

                        # find a non CNOT+Rz gate. Set it a terminate node
                        if p_node.gate_type is None or p_node.gate_type not in \
                                [GateType.cx, GateType.x, GateType.rz] or p_node.flag == p_node.FLAG_VISITED:
                            term_node_set.add(id(p_node))
                            break

                        if p_node.gate_type == GateType.cx:
                            # encounter a new CNOT gate
                            visited_cnot[id(p_node)] = [p_node, 0b00]
                            # if we encounter a CNOT acting on a new qubit, set it as an anchor node
                            o_qubit = p_node.qubit_loc[p_qubit ^ 1]
                            if o_qubit not in anchors:
                                anchors[o_qubit] = p_node
                                anchor_queue.append(o_qubit)
                                # anchor qubit is regarded as initially reachable
                                visited_cnot[id(p_node)][1] |= 1 << (p_qubit ^ 1)

                        c_node = p_node

            # STEP 2: topologically visit the CNOT node in visited_cnot. Calculate reachable bitmasks.
            cnot_queue = deque([node])
            while len(cnot_queue) > 0:
                anchor_node = cnot_queue.popleft()
                for anchor_qubit in anchor_node.qubit_loc:
                    for neighbors in ['predecessors', 'successors']:
                        c_node = anchor_node
                        while True:
                            c_node.flag = c_node.FLAG_VISITED
                            p_node, p_qubit = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]
                            if id(p_node) in term_node_set:
                                break
                            if p_node.gate_type == GateType.cx and id(p_node) in visited_cnot:
                                visited_cnot[id(p_node)][1] |= 1 << p_node.qubit_id[anchor_qubit]
                                if visited_cnot[id(p_node)][1] == 0b11 and p_node.flag == p_node.FLAG_DEFAULT:
                                    p_node.flag = p_node.FLAG_IN_QUE
                                    cnot_queue.append(p_node)
                                break
                            if p_node.flag == p_node.FLAG_VISITED:
                                break
                            c_node = p_node

            # If any of the two qubits of one CNOT is not reachable, it cannot be included
            # in the graph. Set it as terminate node.
            for list_ in visited_cnot.values():
                if list_[1] != 0b11:
                    term_node_set.add(id(list_[0]))

            # STEP 3: calculate boundary by traversing each qubit until we encounter a terminate node
            bound = {}
            visited_node = set()
            for neighbors in ['predecessors', 'successors']:
                bound[neighbors]: List[Tuple[DAG.Node, int]] = [None] * gates.width()
                for anchor_qubit, anchor_node in anchors.items():
                    if id(anchor_node) in term_node_set:
                        continue
                    c_node = anchors[anchor_qubit]
                    while True:
                        visited_node.add(id(c_node))
                        p_node, _ = getattr(c_node, neighbors)[c_node.qubit_id[anchor_qubit]]
                        if id(p_node) in term_node_set:
                            bound[neighbors][anchor_qubit] = (c_node, c_node.qubit_id[anchor_qubit])
                            break
                        c_node = p_node
            yield bound['predecessors'], bound['successors'], len(visited_node)

    def assign_symbolic_phases(self, gates: DAG):
        """
        Assign polarity to undetermined T gates in the circuit. Do it greedily to reduce gate count.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        if not gates.has_symbolic_rz:
            return 0

        # collect all variables in the circuit
        # label -> (variable, expr list)
        var_dict = {}
        for node_ in gates.topological_sort():
            if node_.gate_type == GateType.rz and isinstance(node_.params[0], SymbolicPhase):
                cur_var_dict = node_.params[0].var_dict
                for label in cur_var_dict:
                    var, _ = cur_var_dict[label]
                    if label not in var_dict:
                        var_dict[label] = [var, []]

                    var_dict[label][1].append(node_.params[0])

        ret = 0
        # greedily assign polarity to each variable
        for var, expr_list in var_dict.values():
            var.phase = np.pi / 4
            t_cnt = sum([np.isclose(expr.evaluate(), 0) for expr in expr_list])
            var.phase = -np.pi / 4
            tdg_cnt = sum([np.isclose(expr.evaluate(), 0) for expr in expr_list])
            if t_cnt >= tdg_cnt:
                var.phase = np.pi / 4

            ret += max(t_cnt, tdg_cnt)

        # replace all symbolic param with actual value
        for node_ in gates.topological_sort():
            if node_.gate_type == GateType.rz and isinstance(node_.params[0], SymbolicPhase):
                node_.params[0] = node_.params[0].evaluate()

        # evaluate global phase
        gates.global_phase = gates.global_phase.evaluate()
        gates.has_symbolic_rz = False
        return ret

    def _change_poly_phase(self, node, qubit_, delta, pos_list=None, pos_cnt=None, history=None):
        """
        Change poly phase of `node` on wire `qubit_` by incremental value `delta`.
        Either `pos_list` or `pos_cnt` should be given.
        """

        # remove old value
        cur = node.poly_phase[qubit_]
        if pos_list:
            old = (node, qubit_, cur & 1)
            pos_list[cur >> 1].remove(old)
        if pos_cnt:
            pos_cnt[cur >> 1] -= 1

        # update and add new value
        node.poly_phase[qubit_] ^= delta
        cur = node.poly_phase[qubit_]
        if pos_list:
            new = (node, qubit_, cur & 1)
            if cur >> 1 not in pos_list:
                pos_list[cur >> 1] = []
            pos_list[cur >> 1].append(new)
        if pos_cnt:
            if cur >> 1 not in pos_cnt:
                pos_cnt[cur >> 1] = 0
            pos_cnt[cur >> 1] += 1

        # record the change in history
        if history is not None:
            history.append((node, qubit_, delta))

    def _erase_from_pos_list(self, node, pos_list):
        """
        Maintain floating position list `pos_list` when erasing `node`.
        """
        for qubit_ in range(node.size):
            cur = node.poly_phase[qubit_]
            old = (node, qubit_, cur & 1)
            pos_list[cur >> 1].remove(old)

    def _check_float_pos(self, node, qubit_, phases, pos_list=None, pos_cnt=None):
        """
        Check whether it is valid to change the poly phase of `node` on wire `qubit_`.
        """
        cur = node.poly_phase[qubit_] >> 1
        if cur not in phases or np.isclose(float(phases[cur]), 0):
            return True
        else:
            if pos_list:
                return len(pos_list[cur]) > 1
            else:
                return pos_cnt[cur] > 1

    def _float_cancel_sub_circuit(self, prev_node, succ_node):
        """
        Do float two qubit cancelling in sub circuit.
        """
        # calculate floating positions
        phases = {}
        cur_phases = {}
        pos_list = {}

        # put right bound into a set
        term_set = set()
        for node_, qubit_ in filter(lambda x: x is not None, succ_node):
            term_set.add(id(node_))

        # STEP 1: calculate poly phase of all positions in the circuit
        for qubit_, pack in enumerate(prev_node):
            if pack is not None:
                node_, wire_ = pack
                cur_phases[qubit_] = 1 << (qubit_ + 1)
                pos_list[1 << qubit_] = [(node_, wire_, 0)]

        rz_cnt = 0
        for node_ in list(DAG.topological_sort_sub_circuit(prev_node, succ_node)):
            if node_.gate_type == GateType.cx:
                cur_phases[node_.qubit_loc[1]] = cur_phases[node_.qubit_loc[1]] ^ cur_phases[node_.qubit_loc[0]]
                node_.poly_phase = [cur_phases[node_.qubit_loc[0]], cur_phases[node_.qubit_loc[1]]]
                for qubit_ in range(2):
                    cur = node_.poly_phase[qubit_]
                    if cur >> 1 not in pos_list:
                        pos_list[cur >> 1] = []
                    pos_list[cur >> 1].append((node_, qubit_, cur & 1))

            elif node_.gate_type == GateType.x:
                cur_phases[node_.qubit_loc[0]] = cur_phases[node_.qubit_loc[0]] ^ 1
                node_.poly_phase = [cur_phases[node_.qubit_loc[0]]]
                cur = node_.poly_phase[0]
                if cur >> 1 not in pos_list:
                    pos_list[cur >> 1] = []
                pos_list[cur >> 1].append((node_, 0, cur & 1))

            elif node_.gate_type == GateType.rz:
                # remove rz gates (later we put it back)
                rz_cnt += 1
                sign = -1 if cur_phases[node_.qubit_loc[0]] & 1 else 1
                mono = (cur_phases[node_.qubit_loc[0]] >> 1)
                phases[mono] = sign * node_.params[0] + (phases[mono] if mono in phases else 0)
                node_.erase()

        # STEP 2: cancel cx gates
        cnt = 0

        reachable = DAG.get_sub_circuit_reachable_relation(prev_node, succ_node)
        for cx in list(DAG.topological_sort_sub_circuit(prev_node, succ_node)):
            if cx.gate_type != GateType.cx or cx.flag == cx.FLAG_ERASED:
                continue

            change_history = []
            success = False
            c_ctrl_node, c_ctrl_qubit = cx, 0
            c_targ_node, c_targ_qubit = cx, 1
            while True:
                n_ctrl_node, n_ctrl_qubit = c_ctrl_node.successors[c_ctrl_qubit]
                n_targ_node, n_targ_qubit = c_targ_node.successors[c_targ_qubit]

                # find cancellation
                if id(n_ctrl_node) == id(n_targ_node) and n_ctrl_node.gate_type == GateType.cx and \
                        n_ctrl_qubit == 0 and n_targ_qubit == 1:

                    # cancellation failed
                    if not self._check_float_pos(cx, 1, phases, pos_list=pos_list):
                        break

                    # cancellation succeeded, maintain pos_list and erase nodes
                    success = True
                    self._erase_from_pos_list(n_ctrl_node, pos_list)
                    self._erase_from_pos_list(cx, pos_list)
                    n_ctrl_node.erase()
                    cx.erase()
                    cnt += 2
                    break

                if id(n_ctrl_node) not in term_set and \
                        n_ctrl_node.gate_type == GateType.cx and \
                        n_ctrl_qubit == 0 and \
                        ((id(c_targ_node), c_targ_qubit), id(n_ctrl_node)) not in reachable:
                    # Case A: share control
                    # ---O--O--
                    # ---|--|--
                    # ---X--|--
                    # ------X--
                    c_ctrl_node, c_ctrl_qubit = n_ctrl_node, n_ctrl_qubit

                elif id(n_targ_node) not in term_set and \
                        n_targ_node.gate_type == GateType.cx and \
                        n_targ_qubit == 1 and \
                        ((id(c_ctrl_node), c_ctrl_qubit), id(n_targ_node)) not in reachable and \
                        self._check_float_pos(cx, 1, phases, pos_list=pos_list):
                    # Case B: share target
                    # ---X--X--
                    # ---|--|--
                    # ---O--|--
                    # ------O--
                    self._change_poly_phase(cx, 1, n_targ_node.poly_phase[0], pos_list=pos_list, history=change_history)
                    self._change_poly_phase(n_targ_node, 1, cx.poly_phase[0], pos_list=pos_list, history=change_history)

                    c_targ_node, c_targ_qubit = n_targ_node, n_targ_qubit

                elif id(n_targ_node) not in term_set and n_targ_node.gate_type == GateType.x and \
                        self._check_float_pos(cx, 1, phases, pos_list=pos_list):
                    # Case C: X gate on target
                    # ---O-----
                    # ---|-----
                    # ---X--X--
                    self._change_poly_phase(cx, 1, 1, pos_list=pos_list, history=change_history)
                    self._change_poly_phase(n_targ_node, 0, cx.poly_phase[0], pos_list=pos_list, history=change_history)

                    c_targ_node, c_targ_qubit = n_targ_node, n_targ_qubit
                else:
                    break

            # undo changes if no cancellation found
            if not success:
                for node_, qubit_, delta_ in reversed(change_history):
                    self._change_poly_phase(node_, qubit_, delta_, pos_list=pos_list)

        # put back Rz
        for mono, phase in phases.items():
            if np.isclose(float(phase), 0):
                continue
            rz_cnt -= 1
            c_node, c_qubit, sign = pos_list[mono][0]
            r_node = DAG.Node(Rz(0) & c_node.qubit_loc[c_qubit])
            r_node.params = [phase if sign == 0 else -phase]
            c_node.append(c_qubit, 0, r_node)
        return cnt + rz_cnt

    def float_cancel_two_qubit_gates(self, gates):
        """
        Merge CX gates in the circuit while considering float positions of Rz gates.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        gates.reset_flag()
        gates.set_qubit_loc()
        cnt = 0
        for prev_node, succ_node, node_cnt in list(self._enumerate_cnot_rz_circuit(gates)):
            # the boundary given by enumerate_cnot_rz_circuit is described by internal node of
            # the sub circuit, but PhasePolynomial and replace method need eternal boundary.
            # This conversion is necessary because nodes in eternal boundary may change due to previous iteration.
            for qubit_ in range(gates.width()):
                if prev_node[qubit_] is not None:
                    c_node, c_qubit = prev_node[qubit_]
                    prev_node[qubit_] = c_node.predecessors[c_qubit]
                    c_node, c_qubit = succ_node[qubit_]
                    succ_node[qubit_] = c_node.successors[c_qubit]
            cnt += self._float_cancel_sub_circuit(prev_node, succ_node)
        return cnt

    def _try_float_cancel_sub_circuit(self, prev_node, succ_node):
        """
        Test whether float_cancel_sub_circuit can find cancellation in this sub circuit.
        The difference with `_float_cancel_sub_circuit` is that we do not change the circuit.
        Thus instead of recording every possible floating position, we only record the count of them.
        """

        phases = {}
        cur_phases = {}
        pos_cnt = {}

        # put right bound into a set
        term_set = set()
        for node_, qubit_ in filter(lambda x: x is not None, succ_node):
            term_set.add(id(node_))

        # STEP 1: calculate poly phase of all positions in the circuit
        for qubit_, pack in enumerate(prev_node):
            if pack is not None:
                cur_phases[qubit_] = 1 << (qubit_ + 1)
                pos_cnt[1 << qubit_] = 1

        rz_cnt = 0
        for node_ in list(DAG.topological_sort_sub_circuit(prev_node, succ_node)):
            if node_.gate_type == GateType.cx:
                cur_phases[node_.qubit_loc[1]] = cur_phases[node_.qubit_loc[1]] ^ cur_phases[node_.qubit_loc[0]]
                node_.poly_phase = [cur_phases[node_.qubit_loc[0]], cur_phases[node_.qubit_loc[1]]]
                for qubit_ in range(2):
                    cur = node_.poly_phase[qubit_]
                    if cur >> 1 not in pos_cnt:
                        pos_cnt[cur >> 1] = 0
                    pos_cnt[cur >> 1] += 1

            elif node_.gate_type == GateType.x:
                cur_phases[node_.qubit_loc[0]] = cur_phases[node_.qubit_loc[0]] ^ 1
                node_.poly_phase = [cur_phases[node_.qubit_loc[0]]]
                cur = node_.poly_phase[0]
                if cur >> 1 not in pos_cnt:
                    pos_cnt[cur >> 1] = 0
                pos_cnt[cur >> 1] += 1

            elif node_.gate_type == GateType.rz:
                rz_cnt += 1
                sign = -1 if cur_phases[node_.qubit_loc[0]] & 1 else 1
                mono = (cur_phases[node_.qubit_loc[0]] >> 1)
                phases[mono] = sign * node_.params[0] + (phases[mono] if mono in phases else 0)

        # return True if we can merge some Rz gates
        for phase in phases.values():
            if not np.isclose(float(phase), 0):
                rz_cnt -= 1
        if rz_cnt > 0:
            return True

        # STEP 2: try cx cancellation
        reachable = DAG.get_sub_circuit_reachable_relation(prev_node, succ_node)
        for cx in list(DAG.topological_sort_sub_circuit(prev_node, succ_node)):
            if cx.gate_type != GateType.cx or cx.flag == cx.FLAG_ERASED:
                continue

            change_history = []
            c_ctrl_node, c_ctrl_qubit = cx, 0
            c_targ_node, c_targ_qubit = cx, 1
            while True:
                n_ctrl_node, n_ctrl_qubit = c_ctrl_node.successors[c_ctrl_qubit]
                n_targ_node, n_targ_qubit = c_targ_node.successors[c_targ_qubit]
                if id(n_ctrl_node) == id(n_targ_node) and n_ctrl_node.gate_type == GateType.cx and \
                        n_ctrl_qubit == 0 and n_targ_qubit == 1:
                    if not self._check_float_pos(cx, 1, phases, pos_cnt=pos_cnt):
                        break
                    return True

                if id(n_ctrl_node) not in term_set and n_ctrl_node.gate_type == GateType.rz:
                    # Extra Case 1
                    # --O--Rz--
                    # --|------
                    # --X------
                    c_ctrl_node, c_ctrl_qubit = n_ctrl_node, n_ctrl_qubit
                elif id(n_targ_node) not in term_set and n_targ_node.gate_type == GateType.rz:
                    # Extra Case 2
                    # --O------
                    # --|------
                    # --X--Rz--
                    c_targ_node, c_targ_qubit = n_targ_node, n_targ_qubit
                elif id(n_ctrl_node) not in term_set and \
                        n_ctrl_node.gate_type == GateType.cx and \
                        n_ctrl_qubit == 0 and \
                        ((id(c_targ_node), c_targ_qubit), id(n_ctrl_node)) not in reachable:
                    # Case A: share control
                    # ---O--O--
                    # ---|--|--
                    # ---X--|--
                    # ------X--
                    c_ctrl_node, c_ctrl_qubit = n_ctrl_node, n_ctrl_qubit

                elif id(n_targ_node) not in term_set and \
                        n_targ_node.gate_type == GateType.cx and \
                        n_targ_qubit == 1 and \
                        ((id(c_ctrl_node), c_ctrl_qubit), id(n_targ_node)) not in reachable and \
                        self._check_float_pos(cx, 1, phases, pos_cnt=pos_cnt):
                    # Case B: share target
                    # ---X--X--
                    # ---|--|--
                    # ---O--|--
                    # ------O--
                    self._change_poly_phase(cx, 1, n_targ_node.poly_phase[0], pos_cnt=pos_cnt, history=change_history)
                    self._change_poly_phase(n_targ_node, 1, cx.poly_phase[0], pos_cnt=pos_cnt, history=change_history)

                    c_targ_node, c_targ_qubit = n_targ_node, n_targ_qubit

                elif id(n_targ_node) not in term_set and n_targ_node.gate_type == GateType.x and \
                        self._check_float_pos(cx, 1, phases, pos_cnt=pos_cnt):
                    # Case C: X gate on target
                    # ---O-----
                    # ---|-----
                    # ---X--X--
                    self._change_poly_phase(cx, 1, 1, pos_cnt=pos_cnt, history=change_history)
                    self._change_poly_phase(n_targ_node, 0, cx.poly_phase[0], pos_cnt=pos_cnt, history=change_history)

                    c_targ_node, c_targ_qubit = n_targ_node, n_targ_qubit
                else:
                    break

            for node_, qubit_, delta_ in reversed(change_history):
                self._change_poly_phase(node_, qubit_, delta_, pos_cnt=pos_cnt)

        return False

    def try_float_cancel_two_qubit_gates(self, gates):
        """
        Test whether float_cancel_two_qubit_gates can reduce gate count in the current circuit.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            bool: whether float_cancel_two_qubit_gates can reduce gate count
        """

        gates.reset_flag()
        gates.set_qubit_loc()
        for prev_node, succ_node, node_cnt in list(self._enumerate_cnot_rz_circuit(gates)):
            # the boundary given by enumerate_cnot_rz_circuit is described by internal node of
            # the sub circuit, but PhasePolynomial and replace method need eternal boundary.
            # This conversion is necessary because nodes in eternal boundary may change due to previous iteration.
            for qubit_ in range(gates.width()):
                if prev_node[qubit_] is not None:
                    c_node, c_qubit = prev_node[qubit_]
                    prev_node[qubit_] = c_node.predecessors[c_qubit]
                    c_node, c_qubit = succ_node[qubit_]
                    succ_node[qubit_] = c_node.successors[c_qubit]
            if self._try_float_cancel_sub_circuit(prev_node, succ_node):
                return True

        return False

    def gate_preserving_rewrite(self, gates):
        """
        Reduce gate count by gate preserving rewrite rules.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """

        cnt = 0
        success = True
        while success:
            success = False
            for template in self.gate_preserving_rewrite_template:
                for node in gates.topological_sort():
                    mapping = template.compare((node, -1), dummy_rz=True)
                    if not mapping:
                        continue
                    original, undo_mapping = template.regrettable_replace(mapping)

                    if self.try_float_cancel_two_qubit_gates(gates):
                        # if rewriting enables cancellation else where, do it and restart template matching
                        cnt += self.float_cancel_two_qubit_gates(gates)
                        success = True
                        break

                    # undo replacing if otherwise
                    template.undo_replace(original, undo_mapping)
                if success:
                    break
        return cnt

    def gate_reducing_rewrite(self, gates):
        """
        Reduce gate count by gate reducing rewrite rules.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        cnt = 0
        for template in self.gate_reducing_rewrite_template:
            cnt += template.replace_all(gates) * template.weight
        if cnt:
            cnt += self.float_cancel_two_qubit_gates(gates)
            cnt += self.cancel_two_qubit_gates(gates)
        return cnt

    def float_rotations(self, gates: DAG):
        """
        Reduce gate count by considering float positions of Rz gates.

        Args:
            gates(DAG): DAG of the circuit
        Returns:
            int: Number of gates reduced.
        """
        cnt = 0
        cnt += self.float_cancel_two_qubit_gates(gates)
        cnt += self.gate_preserving_rewrite(gates)
        cnt += self.gate_reducing_rewrite(gates)
        return cnt

    def __init__(self, level='light', optimize_toffoli=True, verbose=False):
        """
        Heuristic optimization of circuits in Clifford + Rz.

        Args:
              level(str): Support 'light' and 'heavy' level. See details in [1]
              optimize_toffoli(bool): whether decompose and optimize ccx/ccz gates into Clifford+rz
              verbose(bool): whether output details of each step

        [1] Nam, Yunseong, et al. "Automated optimization of large quantum
        circuits with continuous parameters." npj Quantum Information 4.1
        (2018): 1-12.
        """
        assert level in self._optimize_routine, Exception(f'unrecognized level {level}')
        self.level = level
        self.verbose = verbose
        self.optimize_toffoli = optimize_toffoli

    def __repr__(self):
        return f'CliffordRzOptimization(level={self.level}, ' \
               f'optimize_toffoli={self.optimize_toffoli})'

    @OutputAligner()
    def execute(self, gates):
        """
        Heuristic optimization of circuits in Clifford + Rz.

        Args:
              gates(Circuit): Circuit to be optimized

        Returns:
            Circuit: The circuit after optimization

        [1] Nam, Yunseong, et al. "Automated optimization of large quantum
        circuits with continuous parameters." npj Quantum Information 4.1
        (2018): 1-12.
        """
        routine = self._optimize_routine[self.level]
        _gates = DAG(gates, self.optimize_toffoli)
        self.parameterize_all(_gates)

        gate_cnt = 0
        round_cnt = 0
        total_time = 0

        while True:
            round_cnt += 1
            if self.verbose:
                print(f'ROUND #{round_cnt}:')

            cnt = 0
            # apply each method in routine
            for step in routine:
                start_time = time.time()
                cur_cnt = getattr(self, self._optimize_sub_method[step])(_gates)
                end_time = time.time()

                if self.verbose:
                    print(f'\t{self._optimize_sub_method[step]}: {cur_cnt} '
                          f'gates reduced, cost {np.round(end_time - start_time, 3)} s')

                cnt += cur_cnt
                total_time += end_time - start_time

            if cnt == 0:
                # assign symbolic t gates for nearly optimized circuit
                start_time = time.time()
                cnt += self.assign_symbolic_phases(_gates)
                end_time = time.time()
                if self.verbose:
                    print(f'\tassign_symbolic_phases: {cnt} '
                          f'gates reduced, cost {np.round(end_time - start_time, 3)} s')

                # stop if nothing can be optimized
                if cnt == 0:
                    break

            gate_cnt += cnt

        self.deparameterize_all(_gates)
        ret = _gates.get_circuit()

        if self.verbose:
            print(f'initially {_gates.init_size} gates, '
                  f'remain {ret.size()} gates, cost {np.round(total_time, 3)} s')
        return ret



import os
import sqlite3


def sql_locked(function):
    def decorator(self, *args, **kwargs):
        self._connect.execute('BEGIN EXCLUSIVE')
        result = function(self, *args, **kwargs)
        self._connect.commit()

        return result

    return decorator


class SQLManger:
    """ Using SQL database to store user information and password. """
    def __init__(self):
        file_path = os.path.dirname(__file__)
        self._connect = sqlite3.connect(f"{file_path}/user_info.db")
        self._connect.isolation_level = "EXCLUSIVE"
        self._cursor = self._connect.cursor()

        try:
            self._create_table()
        except Exception as _:
            pass

    def __del__(self):
        self._cursor.close()
        self._connect.close()

    def clean(self):
        self._cursor.execute("DROP TABLE USER_PASS_MAPPING")
        self._cursor.execute("DROP TABLE USER_STATIC_INFO")

    def _create_table(self):
        # Temp add here, delete after create datebase
        self._cursor.execute('CREATE TABLE USER_PASS_MAPPING(NAME text unique, PASSWORD text)')
        self._cursor.execute(
            'CREATE TABLE USER_STATIC_INFO( \
                NAME text unique, EMAIL text, LEVEL INT, MAX_RUNNING_JOB INT, MAX_STOPPED_JOB INT, GPU_ALLOWENCE BOOL)'
        )

    def validation_password(self, username: str, passwd: str) -> bool:
        # Get Stored User Passwd
        self._cursor.execute(f'SELECT PASSWORD FROM USER_PASS_MAPPING WHERE NAME=\'{username}\'')
        user_passwd = self._cursor.fetchall()[0]

        return user_passwd[0] == passwd

    def get_password(self, username: str) -> str:
        self._cursor.execute(f'SELECT PASSWORD FROM USER_PASS_MAPPING WHERE NAME=\'{username}\'')
        user_passwd = self._cursor.fetchall()[0]

        return user_passwd[0]

    def validate_user(self, username: str) -> bool:
        user_data = self.get_user_info(username)

        return user_data is not None

    @sql_locked
    def update_password(self, username: str, new_passwd: str):
        self._cursor.execute(f"UPDATE USER_PASS_MAPPING SET PASSWORD = \'{new_passwd}\' WHERE NAME = \'{username}\'")

    @sql_locked
    def add_user(self, user_info: dict):
        username = user_info['username']
        passwd = user_info['password']
        email = user_info['email']
        level = 3

        # Add user-passwd into USER_PASS_MAPPING
        self._cursor.execute(f"INSERT INTO USER_PASS_MAPPING VALUES (\'{username}\', \'{passwd}\')")

        # Add user info into database
        max_rjobs = level * 5
        max_sjobs = level * 10
        gpu_allow = True
        self._cursor.execute(
            f"INSERT INTO USER_STATIC_INFO \
            VALUES(\'{username}\', \'{email}\', \'{level}\', \'{max_rjobs}\', \'{max_sjobs}\', \'{gpu_allow}\')"
        )

    def get_user_info(self, username: str) -> tuple:
        """ return (username, email, level, max_rjobs, max_sjobs) """
        self._cursor.execute(f"SELECT * FROM USER_STATIC_INFO WHERE NAME=\'{username}\'")
        user_info = self._cursor.fetchone()

        return user_info

    @sql_locked
    def update_user_email(self, username: str, new_email: str):
        self._cursor.execute(f"UPDATE USER_STATIC_INFO SET EMAIL = \'{new_email}\' WHERE NAME = \'{username}\'")

    @sql_locked
    def update_user_level(self, username: str, new_level: int):
        max_rjobs = new_level * 3
        max_sjobs = new_level * 10

        self._cursor.execute(
            f"UPDATE USER_STATIC_INFO SET \
            LEVEL = \'{new_level}\', MAX_RUNNING_JOB = \'{max_rjobs}\', MAX_STOPPED_JOB = \'{max_sjobs}\' \
            WHERE NAME = \'{username}\'"
        )

    @sql_locked
    def delete_user(self, username: str):
        try:
            # DELETE FROM USER_PASS_MAPPING
            self._cursor.execute(f"DELETE FROM USER_PASS_MAPPING WHERE NAME=\'{username}\'")
            # DELETE FROM USER_STATIC_INFO
            self._cursor.execute(f"DELETE FROM USER_STATIC_INFO WHERE NAME=\'{username}\'")
        except:
            self._connect.rollback()    # RollBack if failure to delete
